% Box - a box for ray tracing (see Shape)
%

classdef Box < Shape
  properties(GetAccess = public, SetAccess = private)
    length = [1; 1; 1];
    halfLength = [1; 1; 1] / 2;
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constructor and Configuration
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function obj = Box(varargin)
      switch nargin
        case 0
          % do nothing
        case 1
          arg = varargin{1};
          if isscalar(arg) && isa(arg, 'Box')
            obj.setTo(arg);
          elseif Vec3.isSingleVector(arg)
            obj.setSize(arg);
          else
            error('Single argument must be a 3-vector or a Box.')
          end
        case 3
          arg1 = varargin{1};
          arg2 = varargin{2};
          arg3 = varargin{3};
          if isscalar(arg1) && isnumeric(arg1) && ...
              isscalar(arg2) && isnumeric(arg2) && ...
              isscalar(arg3) && isnumeric(arg3)
            obj.setSize([arg1; arg2; arg3]);
          else
            error('Argument must be a numeric scalar.')
          end
        otherwise
          error('Constructor requires 0, 1 or 3 arguments.')
      end
    end
    
    function obj = copy(sourceObj)
      obj = Box(sourceObj);
    end
    
    function setTo(obj, sourceObj)
      Shape.setTo(obj, sourceObj);
      obj.length = sourceObj.length;
      obj.halfLength = sourceObj.halfLength;
    end
    
    function setSize(obj, lenArg)
      obj.length = Vec3.returnSingleVector(lenArg);
      obj.halfLength = lenArg / 2;
      obj.maxRadius = Vec3.length(lenArg);
    end
    
    function setCornerPosition(obj, pos)
      pos = Vec3.returnSingleVector(pos);
      Shape.setPosition(obj, pos + obj.halfLenth);
    end
  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Shape Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vol = volume(obj)
      vol = prod(obj.length);
    end
    function area = surfaceArea(obj)
      [x, y, z] = Vec3.xyz(obj.length);
      area = 2 * (x * y + y * z + z * x);
    end
    function area = crossSectionZ(obj)
      [x, y, z] = Vec3.xyz(obj.length);
      area = x * y;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Surface Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vNorm = surfaceNormal(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      if size(pos, 2) == 1
        % which face does pos point toward?
        [val, nDim] = max(abs(pos ./ obj.halfLength));
        
        % vNorm is in that direction
        vNorm = obj.mLocalToGlobal(:, nDim) * sign(pos(nDim));
      else
        % same for many positions...
        vNorm = zeros(size(pos));
        for n = 1:size(pos, 2)
          [val, nDim] = max(abs(pos(:, n) ./ obj.halfLength));
          vNorm(:, n) = obj.mLocalToGlobal(:, nDim) * sign(pos(nDim, n));
        end
      end        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Inside and Outside
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function [isIt, faceNumber] = isOnSurface(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      posT = pos';
      halfLen = obj.halfLength;
      
      % compute face number
      %     0 => not on a face
      % +/- 1 => +/- X
      % +/- 2 => +/- Y
      % +/- 3 => +/- Z
      faceNumber = (abs(posT(:, 1) - halfLen(1)) < Shape.surfaceTolerance) + ...
         2 * (abs(posT(:, 1) + halfLen(1)) < Shape.surfaceTolerance) + ...
         4 * (abs(posT(:, 2) - halfLen(2)) < Shape.surfaceTolerance) + ...
         6 * (abs(posT(:, 2) + halfLen(2)) < Shape.surfaceTolerance) + ...
         8 * (abs(posT(:, 3) - halfLen(3)) < Shape.surfaceTolerance) + ...
        16 * (abs(posT(:, 3) + halfLen(3)) < Shape.surfaceTolerance);
      
      % is it on the surface?
      isIt = faceNumber ~= 0;
    end
    function isIt = isInside(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      % find normalized distance from center
      if size(pos, 2) == 1
        maxNormDist = max(abs(pos ./ obj.halfLength));
      else
        halfLen = repmat(obj.halfLength, 1, size(pos, 2));
        maxNormDist = max(abs(pos ./ halfLen))';
      end
      
      % is the maximum normalized distance less than 1?
      %  if so, it is inside.
      isIt = maxNormDist < 1;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Collision Checking
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function tHit = hitSurface(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      posT = pos';
      velT = vel';
      
      halfLen = obj.halfLength;
      
      % which indexes to use when checking each dimention
      ind = [2 3 1; 3 1 2; 1 2 3];
      
      % loop variables
      tHit = Inf(Nvec, 1);
      tHitIn = zeros(Nvec, 1);
      tHitOut = zeros(Nvec, 1);
      isNewHit = false(Nvec, 1);
      xyHit = zeros(Nvec, 1);
      
      % loop over dimentions
      for n = 1:3
        % next hit time for relevant plane
        nIs = vel(n, :) > 0;
        
        % on the inside of the box
        tHitIn(nIs) = -(posT(nIs, n) - halfLen(n)) ./ velT(nIs, n);
        tHitIn(~nIs) = -(posT(~nIs, n) + halfLen(n)) ./ velT(~nIs, n);
        tHitIn(tHitIn < 0) = Inf;

        % on the outside of the box
        tHitOut(nIs) = -(posT(nIs, n) + halfLen(n)) ./ velT(nIs, n);
        tHitOut(~nIs) = -(posT(~nIs, n) - halfLen(n)) ./ velT(~nIs, n);
        tHitOut(tHitOut < 0) = Inf;
        
        % check to see if this plane crossing is really a hit
        tHitNext = min(tHitIn, tHitOut);
        nIs = tHitNext < tHit;
        if any(nIs)
          isNewHit(nIs) = true;
          isNewHit(~nIs) = false;
          for m = 1:2
            nn = ind(n, m);
            xyHit(nIs) = posT(nIs, nn) + tHitNext(nIs) .* velT(nIs, nn);
            isNewHit(nIs) = isNewHit(nIs) & xyHit(nIs) >= -halfLen(nn) & ...
              xyHit(nIs) <= halfLen(nn);
          end
          % update tHit
          tHit(isNewHit) = tHitNext(isNewHit);
        end
      end
    end
    
    function tHit = hitFromInside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      posT = pos';
      velT = vel';
      
      halfLen = obj.halfLength;
      
      % loop variables
      tHit = Inf(Nvec, 1);
      tHitNext = zeros(Nvec, 1);
      
      % loop over dimentions
      for n = 1:3
        % next hit time for relevant plane
        nIs = vel(n, :) > 0;
        tHitNext(nIs) = -(posT(nIs, n) - halfLen(n)) ./ velT(nIs, n);
        tHitNext(~nIs) = -(posT(~nIs, n) + halfLen(n)) ./ velT(~nIs, n);
        
        % no need to check to see if this plane crossing is really a hit
        %  as in hitFromOutside, since we assume that we are inside!
        
        % update tHit
        isNewHit = isfinite(tHitNext) & tHitNext < tHit;
        tHit(isNewHit) = tHitNext(isNewHit);
      end
    end
    
    function tHit = hitFromOutside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      [pos, vel, Nvec] = Vec3.matchSize(pos, vel);
      posT = pos';
      velT = vel';
      
      halfLen = obj.halfLength;
      
      % which indexes to use when checking each dimention
      ind = [2 3 1; 3 1 2; 1 2 3];
      
      % loop variables
      tHit = Inf(Nvec, 1);
      tHitNext = zeros(Nvec, 1);
      isNewHit = false(Nvec, 1);
      xyHit = zeros(Nvec, 1);
      
      % loop over dimentions
      for n = 1:3
        % next hit time for relevant plane
        nIs = vel(n, :) > 0;
        tHitNext(nIs) = -(posT(nIs, n) + halfLen(n)) ./ velT(nIs, n);
        tHitNext(~nIs) = -(posT(~nIs, n) - halfLen(n)) ./ velT(~nIs, n);
        
        % check to see if this plane crossing is really a hit
        nIs = tHitNext > 0 & tHitNext < tHit;
        if any(nIs)
          isNewHit(nIs) = true;
          isNewHit(~nIs) = false;
          for m = 1:2
            nn = ind(n, m);
            xyHit(nIs) = posT(nIs, nn) + tHitNext(nIs) .* velT(nIs, nn);
            isNewHit(nIs) = isNewHit(nIs) & xyHit(nIs) >= -halfLen(nn) & ...
              xyHit(nIs) <= halfLen(nn);
          end
          % update tHit
          tHit(isNewHit) = tHitNext(isNewHit);
        end
      end
    end
  end
end % classdef
