% Shape - abstract shape class for ray tracing
%
% ---- Public Properties
%   maxRadius
%     largest distance from shape's origin to any point it's surface
%   position
%     global coordinates of shape's origin (local coordinate [0;0;0])
%   orientation
%     shape's z-axis in global coordinates
%   mGlobalToLocal
%     rotational transformation from global to local basis
%
% ---- Public Methods
%   isIt = isOnSurface(pos)
%     is position POS on the surface of this shape?
%   isIt = isInside(pos)
%     is position POS inside of this shape?
%   tHit = hitSurface(pos, vel)
%     time to impact with this shape, moving from POS with velocity VEL
%   tHit = hitFromInside(pos, vel)
%     time to impact with this shape, moving from POS assuming isIn
%   tHit = hitFromOutside(pos, vel)
%     time to impact with this shape, moving from POS assuming ~isIn
%
%   [pos, vel] = transformTo(pos, vel)
%     transform from global to local coordinates
%   setPositionAndOrientation(pos, orient)
%     set the transformation from global to local coordinates

classdef Shape < handle
  properties (Constant)
    surfaceTolerance = 1e-9;
  end
  properties(GetAccess = public, SetAccess = protected)
    maxRadius = Inf;
    position = Vec3.origin;
    orientation = Vec3.zAxis;
    mGlobalToLocal = eye(3);
    mLocalToGlobal = eye(3);
    surfColor = [1 1 1]';
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    % constructor
    function shape = Shape
    end
    
    % instance copy
    function setTo(obj, sourceObj)
      obj.maxRadius = sourceObj.maxRadius;
      obj.position = sourceObj.position;
      obj.orientation = sourceObj.orientation;
      obj.mGlobalToLocal = sourceObj.mGlobalToLocal;
      obj.mLocalToGlobal = sourceObj.mLocalToGlobal;
      obj.surfColor = sourceObj.surfColor;
    end
    
    % transform from global to local coordinates
    function [pos, vel] = transformTo(obj, pos, vel)
      if size(pos, 2) == 1
        pos = pos - obj.position;
      else
        pos = pos - Vec3.copy(obj.position, size(pos, 2));
      end
      pos = obj.mGlobalToLocal * pos;
      if nargin > 2
        vel = obj.mGlobalToLocal * vel;
      end
    end
    
    % set the transformation from global to local coordinates
    function setPosition(obj, pos)
      obj.position = Vec3.returnSingleVector(pos);
    end
  
    % set the transformation from global to local coordinates
    function setOrientation(obj, orient)
      orient = Vec3.returnSingleVector(orient);
      
      [obj.orientation, len] = Vec3.normalize(orient);
      
      % create global to local rotation
      theta = pi * (len - 1);
      [mG2L, mL2G] = Vec3.alignOpZ(obj.orientation);
      [mzG2L, mzL2G] = Vec3.rotationOpZ(theta);

      obj.mGlobalToLocal = mzG2L * mG2L;
      obj.mLocalToGlobal = mL2G * mzL2G;
    end
  
    % set the transformation from global to local coordinates
    function setPositionAndOrientation(obj, pos, orient)
      setPosition(obj, pos);
      setOrientation(obj, orient);
    end
  
    function setColor(obj, cSurf)
      obj.surfColor = Vec3.returnSingleVector(cSurf);
    end
    
    function cSurf = surfaceColor(obj, pos, isLocal)
      cSurf = Vec3.copy(obj.surfColor, size(pos, 2));
    end
    
    function isIt = isOutside(obj, pos, varargin)
      isIt = ~isInside(obj, pos, varargin{:});
    end
    
  end
    
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Abstract Methods
  methods (Abstract)
    obj = copy(sourceObj);
    
    isOn = isOnSurface(obj, pos, isLocal)
    isIn = isInside(obj, pos, isLocal)
    
    tHit = hitSurface(obj, pos, vel, isLocal)
    tHit = hitFromInside(obj, pos, vel, isLocal)
    tHit = hitFromOutside(obj, pos, vel, isLocal)
    
    vNorm = surfaceNormal(obj, pos, isLocal);
    
    vol = volume(obj);
    area = surfaceArea(obj);
    area = crossSectionZ(obj);
  end
  
  methods (Static)
    function posHit = hitPosition(pos, vel, tHit)
      posHit = Vec3.sum(pos, Vec3.prod(tHit, vel));
    end
  end
end % classdef
