% Sphere - a sphere for ray tracing (see Shape)
%

classdef Sphere < Shape
  properties(GetAccess = public, SetAccess = private)
    radius = 1;
    radiusSquared = 1;
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Public Methods
  methods
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constructor and Configuration
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function obj = Sphere(varargin)
      switch nargin
        case 0
          % do nothing
        case 1
          arg = varargin{1};
          if isscalar(arg) && isa(arg, 'Sphere')
            obj.setTo(arg);
          elseif isscalar(arg) && isnumeric(arg)
            obj.setRadius(arg);
          else
            error('Argument must be a numeric scalar, or Sphere object.')
          end
        otherwise
          error('Constructor requires 0 or 1 arguments.')
      end
    end
    
    function obj = copy(sourceObj)
      obj = Sphere(sourceObj);
    end
    
    function setTo(obj, sourceObj)
      Shape.setTo(obj, sourceObj);
      obj.radius = sourceObj.radius;
      obj.radiusSquared = sourceObj.radiusSquared;
    end
    
    function setRadius(obj, radArg)
      obj.radius = radArg;
      obj.radiusSquared = radArg^2;
      obj.maxRadius = radArg;
    end
      
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Shape Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vol = volume(obj)
      vol = (4/3) * pi * obj.radius^3;
    end
    function area = surfaceArea(obj)
      area = 4 * pi * obj.radiusSquared;
    end
    function area = crossSectionZ(obj)
      area = pi * obj.radiusSquared;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Surface Properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function vNorm = surfaceNormal(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      % sphere normal is in radial direction
      vNorm = Vec3.normalize(pos);
      vNorm = obj.mLocalToGlobal * vNorm;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Inside and Outside
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function isOn = isOnSurface(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      l2 = Vec3.lengthSquared(pos);
      isOn = abs(l2 - obj.radiusSquared) < Shape.surfaceTolerance;
    end
    function isIn = isInside(obj, pos, isLocal)
      isLocal = nargin > 2 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        pos = obj.transformTo(pos);
      end
      
      isIn = Vec3.lengthSquared(pos) < obj.radiusSquared;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Collision Checking
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    function tHit = hitSurface(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % prepare quadratic equation parts
      a = Vec3.lengthSquared(vel);
      b = Vec3.dot(pos, vel);
      c = Vec3.lengthSquared(pos) - obj.radiusSquared;
      
      % precompute and eliminate imaginary results
      rtArg = b.^2 - c .* a;
      isHit = rtArg > 0;
      
      if ~any(isHit)
        tHit = Inf(size(a));
        return
      end
      
      % solve
      rt = Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      tPos = (-b + rt) ./ a;
      tNeg = (-b - rt) ./ a;
      
      % find first impact, if any
      isPosOk = tPos > 0;
      isNegOk = tNeg > 0;
      
      tHit = Inf(size(tPos));
      tHit(isPosOk) = tPos(isPosOk);
      tHit(isNegOk) = tNeg(isNegOk);
    end
    
    function tHit = hitFromInside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % prepare quadratic equation parts
      a = Vec3.lengthSquared(vel);
      b = Vec3.dot(pos, vel);
      c = Vec3.lengthSquared(pos) - obj.radiusSquared;
      
      % assume that rtArg > 0
      rtArg = b.^2 - c .* a;
      rt = real(sqrt(rtArg));
      
      % assume that b + rt > 0, so that only one tHit is possible
      tHit = (-b + rt) ./ a;
    end
    
    function tHit = hitFromOutside(obj, pos, vel, isLocal)
      isLocal = nargin > 3 && isLocal;
        
      % transform to local coordinates
      if ~isLocal
        [pos, vel] = obj.transformTo(pos, vel);
      end
      
      % prepare quadratic equation parts
      a = Vec3.lengthSquared(vel);
      b = Vec3.dot(pos, vel);
      c = Vec3.lengthSquared(pos) - obj.radiusSquared;
      
      % precompute and eliminate imaginary results
      %    b < 0    => approaching the origin
      %    c >  0   => outside the sphere
      %    rtArg > 0 => path interescts infinite cylinder
      rtArg = b.^2 - c .* a;
      isHit = b < 0 & c > 0 & rtArg > 0;
      
      if ~any(isHit)
        tHit = Inf(size(a));
        return
      end
      
      rt = -Inf(size(rtArg));
      rt(isHit) = real(sqrt(rtArg(isHit)));
      
      % assume that b + rt < 0, so that only one tHit is possible
      tHit = -(b + rt) ./ a;
      
      % check
      if any(tHit < 0)
        error('Negative hit time')
      end
    end
  end % methods
end % classdef
