% [vVel, vMB] = getFluxVel(temp, mass, vNorm, N)
%  temp = temperature in K
%  mass = molecular mass in atomic mass units
%  vNorm = surface normal (velocity will be positive in this direction)
%  N = number of vectors to generate (default = 1)
%
% Flux velocity is through a suface with the specified
% surface normal.  The normal velocity comes from a
% Rayleigh distribution, such that this represents the
% velocity distribution of particles which pass through
% the surface per unit time, assuming the particle velocity
% distribution is Maxwell-Boltzman.
% (see getMaxBoltzVel, getFluxVelNorm)
%
% Rayleigh distribution:
% P(v) / dv = exp(-(v/vp)^2) * (2 * v / vp^2)


function [vVel, vMB] = getFluxVel(temp, mass, vNorm, N)
  
  % set default N
  if nargin < 4
    N = 1;
  end
  
  % start with MaxBoltz velocities, and velocity scale
  [vVelMB, vMB] = getMaxBoltzVel(temp, mass, N);

  % normal velocity component with Rayleigh distribution
  velNorm = getFluxVelNorm(temp, mass, N);

  % change normal component in vVel
  velNormMB = Vec3.dot(vNorm, vVelMB);
  vVel = vVelMB + Vec3.prod(velNorm - velNormMB, vNorm);
  
end
