% [vVel, vMB] = getMaxBoltzVel(temp, mass, N)
%  temp = temperature in K
%  mass = molecular mass in atomic mass units
%  N = number of vectors to generate
%
% vVel = velocity vectors (3xN)
% vMB = characteristic velocity scale
%
% Output vectors are from the Maxwell-Boltzman
% velocity distribution for particles in a gas.
%
% Maxwell-Boltzman distribution:
% P(v) / dv = exp(-(v/vMB)^2) / (sqrt(pi) * vMB)

function [vVel, vMB] = getMaxBoltzVel(temp, mass, N)
  
  % set default N
  if nargin < 3
    N = 1;
  end
  
  % constants
  ATOMIC_MASS = 1.66053886e-27;           % kg/au
  BOLTZMANN_CONST = 1.38066e-23;          % (kg m^2/s^2) / K

  % characteristic velocity
  vMB = sqrt(2 * (BOLTZMANN_CONST / ATOMIC_MASS) * (temp / mass));
  
  % generate velocity components
  %   P(v) / dv = exp(-(v/vMB)^2) / (sqrt(pi) * vMB)
  vVel = vMB * randn(3, N) / sqrt(2);
  
end
