% plotPressure(p, rslt, nPlot)
%
% Example:
% subplot(2, 1, 1)
% plotWallStats(pFlat, rsltFlat)
% title('Flat Box')
%
% subplot(2, 1, 2)
% plotWallStats(pLong, rsltLong)
% title('Tall Box')

function [fspec, f] = plotPressure(p, rslt, Ntotal, nPlot)
  
  
  % pressure
  ATOMIC_MASS = 1.66053886e-27;           % kg/au
  BOLTZMANN_CONST = 1.38066e-23;          % (kg m^2/s^2) / K

  mass = ATOMIC_MASS * p.mass;
  kBT = BOLTZMANN_CONST * p.temp;
  
  % number density
  volume = p.shapes{1}.volume;
  for n = 2:numel(p.shapes)
    volume = volume - p.shapes{n}.volume;
  end
  rho0 = 1e14;  % nominal number density, at 3e-9 torr [1/m^3]
  rho = Ntotal / volume;
  rhoR = rho0 / rho;
  
  
  % characteristic velocity
  [vVel, vMB] = getMaxBoltzVel(p.temp, p.mass);

  % free flow gas mean velocity and pressure
  vzFree = sqrt(pi / 4) * vMB;
  pzFree = rho0 * kBT / 2;
  
  % free flow gas force noise level for a unit area
  fzFree = sqrt(8 * kBT * rho0 * sqrt(kBT * mass));

  % impact normal velocity distributions
  dtBin = p.dtBin;
  tBin = (dtBin:dtBin:p.tEnd)';

  dvBin = p.dvBin;
  vBin = (dvBin:dvBin:p.vMax);
  
  % loop over results
  area = zeros(numel(nPlot), 1);
  areaZ = zeros(numel(nPlot), 1);
  impactRate = zeros(numel(nPlot), 1);
  pressure = zeros(numel(nPlot), 1);
  fz = zeros(rslt(1).numberTimeBins - 1, numel(nPlot));
  for n = 1:numel(nPlot)
    nn = nPlot(n);
    
    % time area product
    areaZ(n) = 2 * p.shapes{nn}.crossSectionZ;
    timeArea = p.tEnd * p.shapes{nn}.surfaceArea;
    
    % impact counter
    impactRate(n) = sum(rslt(nn).velNormDist) / timeArea;

    % pressure
    pressure(n) = sum(rslt(nn).velNormDist .* vBin) * mass / timeArea;
    
    % force, remove overflow bin
    fz(:, n) = rslt(nn).forceVsTime(3, 1:end-1);
  end
  
  % sanity check
  disp('Normalized pressure (should be 1)')
  pErr = pressure * rhoR / pzFree;
  disp(pErr')
  
  disp('Impact rate per unit area / rho')
  disp(impactRate' / rho)
  
  % make spectra
  nfft = 4 * 1024;
  mywin = hann(nfft);
  pspec = zeros(nfft / 2 + 1, numel(nPlot));
  for n = 1:numel(nPlot)
    [pspec(:, n), f] = pwelch(fz(:, n), mywin, nfft-1024, nfft, 1/dtBin);
  end
  
  % convert to amplitude spectrum, normalized by relative number density
  fspec = sqrt(rhoR * pspec);
  
  % plot
  clf
  pspecFree = ones(numel(f), 1) * fzFree;
  fspecFree = pspecFree * sqrt(areaZ)';
  %semilogx(f, [fspec, pspecFree]);
  loglog(f, fspec);
  hold on
  loglog(f, fspecFree, '--');
  grid on
  hold off
  xlabel('frequency [Hz]')
  ylabel('force noise [N/\surd Hz]')
end
